package net.es.oscars.policy;

import java.rmi.RemoteException;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.List;
import java.util.TimeZone;

import net.es.oscars.policy.config.ConfigUtil;
import net.es.oscars.policy.database.Operator;
import net.es.oscars.policy.database.Rule;
import net.es.oscars.policy.database.RuleDAO;

import org.apache.log4j.Logger;

import javax.xml.namespace.NamespaceContext;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.DOMException;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

public class RuleEvaluator {
    private Logger log;
    private String dbname;
    private NamespaceContext ctx;
    
    public RuleEvaluator(String dbname){
        this.log = Logger.getLogger(this.getClass());
        this.dbname = dbname;
        this.ctx = this.getNamespaceContext();
    }
    
    public String evaluate(Rule rule, String action, List<Element[]> resources) throws RemoteException {
        this.log.debug("evaluate.start");
        if(rule == null){
            throw new RemoteException("Encountered empty rule");
        }
        String lType = rule.getLType().getName();
        String reason = null;
        if(lType.startsWith("XPATH_")){
            reason = this.evaluateXPath(rule, resources);
        }else if("SPECIAL_VAR".equals(lType)){
            reason = this.evaluateSpecialVar(rule, resources);
        }else if("ACTION".equals(lType)){
            reason = this.evaluateString(rule, action);
        }else if("RULE".equals(lType)){
            reason = this.evaluateRule(rule, action, resources);
        }else if("*".equals(lType)){
            ;//allow anything
        }else{
            throw new RemoteException("Rule contains invalid LTYPE");
        }
        
        this.log.debug("evaluate.end");
        return reason;
    }

    private String evaluateXPath(Rule rule, List<Element[]> resources) throws RemoteException {
        this.log.debug("evaluateXPath.start");
        String reason = null;
        Operator op = rule.getOperator();
        String xpathType = rule.getLType().getName().split("_")[1];
        if(!(xpathType.equals(op.getRType().getName()) || 
                (xpathType.equals("NUMBER") && "DATE".equals(op.getRType().getName()))|
                xpathType.equals("NODE") || 
                xpathType.equals("LIST") || 
                xpathType.equals("ENUM"))){
            throw new RemoteException("RTYPE must be " + xpathType);
        }
        
        try {
            XPath xpath = XPathFactory.newInstance().newXPath();
            xpath.setNamespaceContext(this.ctx);
            
            boolean elemFound = false;
            for(Element[] resource : resources){
                for(Element resourceElem : resource){
                    if(xpathType.equals("NODE")){
                        Node elem = (Node)xpath.evaluate(rule.getLValue(), resourceElem,XPathConstants.NODE);
                        if(elem != null && elem.getTextContent() != null){
                            reason = this.evaluateNode(rule, elem);
                            elemFound = true;
                            break;
                        }
                    }else if(xpathType.equals("LIST")){
                        NodeList elemList = (NodeList)xpath.evaluate(rule.getLValue(), resourceElem,XPathConstants.NODESET);
                        if(elemList != null && elemList.getLength() > 0){
                            reason = this.evaluateList(rule, elemList);
                            elemFound = true;
                            break;
                        }
                    }else if(xpathType.equals("NUMBER")){
                        String numStr = (String)xpath.evaluate(rule.getLValue(), resourceElem,XPathConstants.STRING);
                        if(numStr != null){
                            this.log.debug("numStr="+numStr);
                            reason = this.evaluateNumber(rule, numStr);
                            elemFound = true;
                            break;
                        }
                    }else if(xpathType.equals("STRING")){
                        String str = (String)xpath.evaluate(rule.getLValue(), resourceElem,XPathConstants.STRING);
                        if(str != null){
                            reason = this.evaluateString(rule, str);
                            elemFound = true;
                            break;
                        }
                    }else if(xpathType.equals("BOOLEAN")){
                        Boolean bool = (Boolean)xpath.evaluate(rule.getLValue(), resourceElem,XPathConstants.BOOLEAN);
                        if(bool != null){
                            reason = this.evaluateBoolean(rule, bool);
                            elemFound = true;
                            break;
                        }
                    }else if(xpathType.equals("ENUM")){
                        NodeList elemList = (NodeList)xpath.evaluate(rule.getLValue(), resourceElem,XPathConstants.NODESET);
                        if(elemList != null && elemList.getLength() > 0){
                            reason = this.evaluateEnum(rule, elemList);
                            elemFound = true;
                            break;
                        }
                    }
                }
                if(elemFound){ 
                    break; 
                }
            }
            if(!elemFound){
                return "Can't find element " + rule.getLValue();
            }
        } catch (XPathExpressionException e) {
            this.log.error(e.getMessage());
            throw new RemoteException("Unable to evaluate rule");
        }
        
        this.log.debug("evaluateXPath.end");
        return reason;
    }

    private String evaluateList(Rule rule, NodeList elemList) throws NumberFormatException, RemoteException, DOMException {
        this.log.debug("evaluateList.start");
        for(int i = 0; i < elemList.getLength(); i++){
            String reason = this.evaluateNode(rule, elemList.item(i));
            if(reason != null){
                return reason;
            }
        }
        this.log.debug("evaluateList.end");
        return null;
    }
    
    private String evaluateEnum(Rule rule, NodeList elemList) throws NumberFormatException, RemoteException, DOMException {
        this.log.debug("evaluateEnum.start");
        for(int i = 0; i < elemList.getLength(); i++){
            String reason = this.evaluateNode(rule, elemList.item(i));
            if(reason == null){
                return null;
            }
        }
        this.log.debug("evaluateEnum.end");
        return "No fields in " + rule.getLValue() + " match constraints";
    }


    private String evaluateNode(Rule rule, Node elem) throws NumberFormatException, RemoteException, DOMException {
        this.log.debug("evaluateNode.start");
        String reason = null;
        Operator op = rule.getOperator();
        
        if(op.getRType().getName().equals("NUMBER")){
            reason = this.evaluateNumber(rule, elem.getTextContent());
        }else if(op.getRType().getName().equals("STRING")){
            reason = this.evaluateString(rule, elem.getTextContent());
        }else if(op.getRType().getName().equals("BOOLEAN")){
            reason = this.evaluateBoolean(rule, "1".equals(elem.getTextContent()));
        }else{
            throw new RemoteException("Invalid operator for evaluating node");
        }
        this.log.debug("evaluateNode.end");
        return reason;
    }

    private String evaluateBoolean(Rule rule, Boolean bool) throws RemoteException {
        this.log.debug("evaluateBoolean.start");
        String reason = null;
        Operator op = rule.getOperator();
        if(!"BOOLEAN".equals(op.getRType().getName())){
            throw new RemoteException("RTYPE must be a BOOLEAN");
        }
        
        if("==".equals(op.getName())){
            if(bool != "1".equals(rule.getRValue())){
                reason = rule.getLValue() + " is not " + "1".equals(rule.getRValue());
            }
        }else if("!=".equals(op.getName())){
            if(bool == "1".equals(rule.getRValue())){
                reason = rule.getLValue() + " equals " + "1".equals(rule.getRValue());
            }
        }
        
        this.log.debug("evaluateBoolean.end");
        return reason;
    }

    private String evaluateString(Rule rule, String value) throws RemoteException {
        this.log.debug("evaluateString.start");
        String reason = null;
        Operator op = rule.getOperator();
        if(!"STRING".equals(op.getRType().getName())){
            throw new RemoteException("RTYPE must be a STRING");
        }
        
        if("==".equals(op.getName())){
            this.log.debug("op is ==");
            if(!value.equals(rule.getRValue())){
                reason = value + " does not equal " + rule.getRValue();
            }
        }else if("!=".equals(op.getName())){
            this.log.debug("op is !=");
            if(value.equals(rule.getRValue())){
                reason =  value + " equals " + rule.getRValue();
            }
        }else if("contains".equals(op.getName())){
            this.log.debug("op is contains");
            if(!value.contains(rule.getRValue())){
                reason =  value + " does not contain " + rule.getRValue();
            }
        }else if("startsWith".equals(op.getName())){
            this.log.debug("op is startsWith");
            if(!value.startsWith(rule.getRValue())){
                reason =  value + " does not start with " + rule.getRValue();
            }
        }else if("endsWith".equals(op.getName())){
            this.log.debug("op is endsWith");
            if(!value.endsWith(rule.getRValue())){
                reason =  value + " does not end with " + rule.getRValue();
            }
        }
        this.log.debug("evaluateString.end");
        
        return reason;
    }
    private String evaluateNumber(Rule rule, String numStr) throws RemoteException{
        String[] ranges = numStr.split(",");
        for(String range : ranges){
            for(String num : range.split("-")){
                double d = 0;
                try{
                    d = Double.parseDouble(num);
                }catch(Exception e){
                    return "Invalid number or range '" + numStr + "'";
                }
                String reason = this.evaluateNumber(rule, d);
                if(reason != null){
                    return reason;
                }
            }
        }
        
        return null;
    }
    private String evaluateNumber(Rule rule, Double value) throws RemoteException {
        this.log.debug("evaluateNumber.start");
        String reason = null;
        Operator op = rule.getOperator();
        double rValue = 0;
        
        //verify value is number
        if(value.isNaN()){
            return (rule.getLValue() + " is not a number.");
        }
        
        if("NUMBER".equals(op.getRType().getName())){
            try{
                rValue = Double.parseDouble(rule.getRValue());
            }catch(Exception e){
                this.log.error(e.getMessage());
                throw new RemoteException("Rule conatins non-numeric rvalue");
            }
        }else if("DATE".equals(op.getRType().getName()) &&
                (rule.getRValue().startsWith("+")||rule.getRValue().startsWith("1"))){
            //relative date
            rValue = this.parseRelativeDate(rule.getRValue());
        }else if("DATE".equals(op.getRType().getName())){
            //absolute date
            rValue = this.evaluateAbsoluteDate(value.longValue(), rule.getRValue());
        }else{
            throw new RemoteException("RTYPE must be a NUMBER or DATE.");
        }
        
        if("==".equals(op.getName())){
            if(value != rValue){
                reason = value + " does not equal " + rule.getRValue();
            }
        }else if("!=".equals(op.getName())){
            if(value == rValue){
                reason = value + " equals " + rule.getRValue();
            }
        }else if("<".equals(op.getName())){
            if(value >= rValue){
                reason = value + " greater than or equal to " + rule.getRValue();
            }
        }else if(">".equals(op.getName())){
            if(value <= rValue){
                reason = value + " less than or equal to " + rule.getRValue();
            }
        }else if("<=".equals(op.getName())){
            if(value > rValue){
                reason = value + " greater than " + rule.getRValue();
            }
        }else if(">=".equals(op.getName())){
            if(value < rValue){
                reason = value + " less than " + rule.getRValue();
            }
        }
        this.log.debug("evaluateNumber.end");
        return reason;
    }
    
    private String evaluateSpecialVar(Rule rule, List<Element[]> resources) throws RemoteException {
        if("current_time".equals(rule.getLValue())){
            return this.evaluateNumber(rule, new Double(System.currentTimeMillis()/1000));
        }
        
        return "Unsupported special variable in rule set " + rule.getLValue();
    }
    
    private String evaluateRule(Rule rule, String action, List<Element[]> resources) throws RemoteException {
        this.log.debug("evaluateRule.start");
        String reason = null;
        Operator op = rule.getOperator();
        if(!"RULE".equals(op.getRType().getName())){
            throw new RemoteException("RTYPE must be a RULE");
        }
        
        RuleDAO ruleDAO = new RuleDAO(this.dbname);
        Rule rule1 = ruleDAO.queryByParam("id", rule.getLValue());
        Rule rule2 = ruleDAO.queryByParam("id", rule.getRValue());
        this.log.debug("rule1="+rule.getLValue());
        String rule1Reason = this.evaluate(rule1, action, resources);
        this.log.debug("rule2="+rule.getRValue());
        String rule2Reason = this.evaluate(rule2, action, resources);
        
        if("AND".equals(op.getName())){
            if(rule1Reason != null && rule2Reason != null){
                reason = rule1Reason + "\n" + rule2Reason;
            }else if(rule1Reason != null){
                reason = rule1Reason;
            }else if(rule2Reason != null){
                reason = rule2Reason;
            }
        }else if("OR".equals(op.getName())){
            if(rule1Reason != null && rule2Reason != null){
                reason = rule1Reason + "\n" + rule2Reason;
            }
        }
        this.log.debug("evaluateRule.end");
        return reason;
    }

    public NamespaceContext getNamespaceContext(){
        NamespaceContext ctx = new NamespaceContext() {
            public String getNamespaceURI(String prefix) {
                if(ConfigUtil.getNamespaceMap().containsKey(prefix)){
                    return ConfigUtil.getNamespaceMap().get(prefix);
                }
                return null;
            }

            public String getPrefix(String namespaceURI) {
                return null;
            }

            public Iterator<String> getPrefixes(String namespaceURI) {
                return null;
            }
        };
        
        return ctx;
    }
    
    private double parseRelativeDate(String dateValue) {
        // TODO Auto-generated method stub
        char op = dateValue.charAt(0);
        String[] parts = dateValue.substring(1).split(" ");
        GregorianCalendar cal = new GregorianCalendar();
        int units = GregorianCalendar.SECOND;
        int amount = Integer.parseInt(parts[0]);
        
        long origTime = cal.getTimeInMillis();
        //get units
        if("minutes".equals(parts[1])){
            units = GregorianCalendar.MINUTE;
        }else if("hours".equals(parts[1])){
            units = GregorianCalendar.HOUR_OF_DAY;
        }else if("days".equals(parts[1])){
            units = GregorianCalendar.DAY_OF_YEAR;
        }else if("weeks".equals(parts[1])){
            units = GregorianCalendar.WEEK_OF_YEAR;
        }else if("months".equals(parts[1])){
            units = GregorianCalendar.MONTH;
        }else if("years".equals(parts[1])){
            units = GregorianCalendar.YEAR;
        }
        
        //perform op
        if(op == '+'){
            cal.add(units, amount);
        }else{
            cal.add(units, -1*amount);
        }
        this.log.info("Change: " + (cal.getTimeInMillis()-origTime));
        return cal.getTimeInMillis()/1000;
    }
    
    private double evaluateAbsoluteDate(long lDate, String rDateStr) throws RemoteException {
        GregorianCalendar lCal = new GregorianCalendar();
        lCal.setTimeZone(TimeZone.getTimeZone("GMT-0"));
        lCal.setTimeInMillis(lDate*1000);
        
        //Get calendar parts
        int[] fields = new int[6];
        fields[0] = GregorianCalendar.YEAR;
        fields[1] = GregorianCalendar.MONTH;
        fields[2] = GregorianCalendar.DAY_OF_MONTH;
        fields[3] = GregorianCalendar.HOUR_OF_DAY;
        fields[4] = GregorianCalendar.MINUTE;
        fields[5] = GregorianCalendar.SECOND;
        
        //Get provided date parts
        GregorianCalendar rCal = new GregorianCalendar();
        rCal.setTimeZone(TimeZone.getTimeZone("GMT-0"));
        String[] strDateArr = new String[6];
        String[] dateTimeArr = rDateStr.split(" ");
        System.arraycopy(dateTimeArr[0].split("-"), 0, strDateArr, 0, 3);
        System.arraycopy(dateTimeArr[1].split(":"), 0, strDateArr, 3, 3);
        for(int i = 0; i < strDateArr.length; i++){
            if("*".equals(strDateArr[i])){
                rCal.set(fields[i], lCal.get(fields[i]));
            }else{
                try{
                    if(fields[i] == GregorianCalendar.MONTH){
                        //month starts at 0 so adjust
                        rCal.set(fields[i], (Integer.parseInt(strDateArr[i]) - 1));
                    }else{
                        rCal.set(fields[i], Integer.parseInt(strDateArr[i]));
                    }
                }catch(Exception e){
                    throw new RemoteException("Error parsing date: " + strDateArr[i] + " is not a number");
                }
            }
        }
        this.log.debug("lcal="+lCal);
        this.log.debug("rcal="+rCal);
        return rCal.getTimeInMillis()/1000;
    }
}
