package net.es.oscars.policy.axis2;

import java.util.HashMap;

import org.apache.axiom.om.OMElement;
import org.apache.axis2.client.Options;
import org.apache.axis2.client.ServiceClient;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.context.ConfigurationContextFactory;
import org.apache.axis2.databinding.types.URI;
import org.apache.axis2.util.XMLUtils;
import org.apache.log4j.Logger;
import org.opensaml.saml2.core.Assertion;

import net.es.oscars.client.security.KeyManagement;
import net.es.oscars.policy.axis2.service.CheckPolicyFaultMessage;
import net.es.oscars.policy.axis2.service.OSCARSPolicyStub;
import net.es.oscars.policy.axis2.types.CheckPolicyRequestType;
import net.es.oscars.policy.axis2.types.RequesterListType;
import net.es.oscars.policy.axis2.types.RequesterType;
import net.es.oscars.policy.axis2.types.RequesterTypeSequence_type0;
import net.es.oscars.policy.axis2.types.ResourceType;
import net.es.oscars.policy.axis2.types.SubjectAuthenticationType;
import net.es.oscars.policy.client.CheckPolicyRequest;
import net.es.oscars.policy.client.CheckPolicyResponse;
import net.es.oscars.policy.client.PolicyClientException;
import net.es.oscars.policy.client.Requester;

/**
 * Client that uses Axis2 to sent and receive checkPolicy messages. Requires resources of
 * type OMElement. 
 * 
 * @author Andrew Lake (alake@internet2.edu)
 *
 */
public class Axis2PolicyClient{
    protected HashMap<String, String> configuration;
    protected Logger log;
    
    final private long DEFAULT_TIMEOUT = 30000; //set to 30 seconds
    final static public String REPO_CONFIG = "repo";
    final static public String AXIS2_CONFIG = "axis2";
    final static public String TIMEOUT_CONFIG = "timeout";
    
    public Axis2PolicyClient(){
        this.configuration = new HashMap<String, String>();
        this.log = Logger.getLogger(this.getClass());
    }
    
    /**
     * Sends a checkPolicy request to the given URL
     * @param url the URL of the policy service
     * @param request the request to send
     * @return the response from the policy server
     * @throws PolicyClientException
     */
    public CheckPolicyResponse checkPolicy(String url, CheckPolicyRequest<Requester,OMElement> request)
            throws PolicyClientException {
        this.log.info("checkPolicy.start");
        //Set up environment
        CheckPolicyResponse response = new CheckPolicyResponse();
        KeyManagement.setKeyStore(this.configuration.get(REPO_CONFIG));
        OSCARSPolicyStub stub = null;
        ConfigurationContext configContext = null;
        try {
            configContext = ConfigurationContextFactory
                    .createConfigurationContextFromFileSystem(
                            this.configuration.get(REPO_CONFIG), 
                            this.configuration.get(AXIS2_CONFIG));
            stub = new OSCARSPolicyStub(configContext, url);
            //Set timeout
            long timeout = DEFAULT_TIMEOUT;
            if(this.configuration.containsKey("timeout")){
                timeout = Long.parseLong(this.configuration.get(TIMEOUT_CONFIG));
                this.log.debug("timeout=" + timeout);
            }
            ServiceClient sc = stub._getServiceClient();
            Options opts = sc.getOptions();
            opts.setTimeOutInMilliSeconds(timeout); 
            sc.setOptions(opts);
            stub._setServiceClient(sc);
            
            //Call service
            net.es.oscars.policy.axis2.types.CheckPolicyResponse
                axis2Response = stub.checkPolicy(this.convertRequest(request));
            
            //Set response
            response.setAllow(axis2Response.getCheckPolicyResponse().getAllow());
            response.setReason(axis2Response.getCheckPolicyResponse().getReason());
            
            //clean up after axis2
            this.cleanUp(configContext);
        } catch (CheckPolicyFaultMessage e) {
            this.cleanUp(configContext);
            throw new PolicyClientException(e.getFaultMessage().getMessage());
        } catch(Exception e){
            this.cleanUp(configContext);
            throw new PolicyClientException(e.getMessage());
        }
        
        return response;
    }
    
    /**
     * Converts a CheckPolicy request to an Axis2 CheckPolicyRequest
     * @param request the request to convert
     * @return the converted request
     * @throws Exception
     */
    private net.es.oscars.policy.axis2.types.CheckPolicyRequest convertRequest(
            CheckPolicyRequest<Requester,OMElement> request) throws Exception {
        net.es.oscars.policy.axis2.types.CheckPolicyRequest axis2Request= 
            new net.es.oscars.policy.axis2.types.CheckPolicyRequest();
        CheckPolicyRequestType axis2RequestType = new CheckPolicyRequestType();
        
        //set Requesters
        RequesterListType requesterList = new RequesterListType();
        int sequenceNumber = 1;
        for(Requester requester : request.getRequesters()){
            //Marshall requester so SAML gets mapped to DOM
            requester.marshall();
            
            //set Subject
            RequesterType axis2Requester = new RequesterType();
            axis2Requester.setSequenceNumber(sequenceNumber);
            axis2Requester.setExtraElement(XMLUtils.toOM(requester.getSubject().getDOM()));
            
            //set SubjectAuthentication
            if(requester.getSubjectAuthentication() != null){
                SubjectAuthenticationType subjAuthType = new SubjectAuthenticationType();
                subjAuthType.setExtraElement(XMLUtils.toOM(requester.getSubjectAuthentication().getDOM()));
                axis2Requester.setSubjectAuthentication(subjAuthType);
            }
            
            //Set Assertions
            if(requester.getAssertions() != null && !requester.getAssertions().isEmpty()){
                RequesterTypeSequence_type0 axis2Assertions = new RequesterTypeSequence_type0();
                for(Assertion assertion : requester.getAssertions()){
                    axis2Assertions.addExtraElement(XMLUtils.toOM(assertion.getDOM()));
                }
                axis2Requester.setRequesterTypeSequence_type0(axis2Assertions);
            }
            
            //increase Requester sequenceNumber
            requesterList.addRequester(axis2Requester);
            sequenceNumber++;
        }
        axis2RequestType.setRequesters(requesterList);
        
        //set Action
        URI actionUri = new URI(request.getAction());
        axis2RequestType.setAction(actionUri);
        
        //set Resources
        for(OMElement resource : request.getResources()){
            ResourceType resourceType = new ResourceType();
            resourceType.addExtraElement(resource);
            axis2RequestType.addResource(resourceType);
        }
        
        //add to message
        axis2Request.setCheckPolicyRequest(axis2RequestType);
        
        return axis2Request;
    }
    
    /**
     * Creates a request with resources of type OMElement. OMElements are the 
     * Axis2 DOM types and so any Axis2 type can be a resource. 
     * 
     * @return a new request with empty parameters
     */
    public CheckPolicyRequest<Requester,OMElement> createRequest() {
        return new CheckPolicyRequest<Requester,OMElement>();
    }

    /**
     * The configuration parameters for this client. Currently it supports
     * setting the repo location (Axis2PolicyClient.REPO_CONFIG), the axis2.xml
     * file (Axis2PolicyClient.AXIS2_CONFIG) and the timeout (Axis2PolicyClient.TIMEOUT_CONFIG).
     * 
     * @return a HashMap with the option names as keys and their current setting as the value
     */
    public HashMap<String, String> getConfiguration() {
        return configuration;
    }
    
    /**
     * Prevents memory leak in axis2 client
     *
     * @param configContext an Axis2 ConfigurationContext to delete
     */
    private void cleanUp(ConfigurationContext configContext){
        if(configContext == null){
            return;
        }
        
        try{
            configContext.terminate();
        }catch(Exception e){
            this.log.warn("Unable to clean-up policy engine. Watch for memory leaks: " + e.getMessage());
        }
    }
}
