package net.es.oscars.policy.client;

import java.io.Serializable;
import java.util.List;

import org.opensaml.SAMLSubject;
import org.opensaml.saml2.core.Assertion;
import org.opensaml.saml2.core.NameID;
import org.opensaml.saml2.core.Subject;
import org.opensaml.saml2.core.SubjectConfirmation;
import org.opensaml.saml2.core.impl.AssertionMarshaller;
import org.opensaml.saml2.core.impl.NameIDBuilder;
import org.opensaml.saml2.core.impl.NameIDMarshaller;
import org.opensaml.saml2.core.impl.SubjectBuilder;
import org.opensaml.saml2.core.impl.SubjectConfirmationBuilder;
import org.opensaml.saml2.core.impl.SubjectMarshaller;
import org.opensaml.xml.ConfigurationException;
import org.opensaml.xml.io.MarshallingException;

/**
 * Represents an entity that requests some action on a resource. Information about the
 * requester is expressed using SAML2 types.
 * 
 * @author Andrew Lake (alake@internet2.edu)
 *
 */
public class Requester implements Serializable{
    private Subject subject;
    private NameID subjectAuthentication;
    private List<Assertion> assertions;
    
    final public static String OSCARS_LOGIN_SUBJECT = "urn:dcn:oscars:loginName";
    
    /**
     * Bootstraps OpenSAML if it has not already been bootstrapped
     * 
     * @throws PolicyClientException
     */
    public Requester() throws PolicyClientException{        
        try {
            SAMLBootstrapper.bootstrap();
        } catch (ConfigurationException e) {
            throw new PolicyClientException(e.getMessage());
        }
    }
    
    /**
     * @return a SAML subject identifying the requester
     */
    public Subject getSubject() {
        return this.subject;
    }
    
    /**
     * Sets the SAML subject that identifies the requester. It will be set to 
     * type X.509 if a comma-delimited name is given and an OSCARS login otherwise.
     * The qualifier is optional and may be set to null. It should only be used for OSCARS
     * login types.
     * 
     * @param name the name of the subject expressed as an X.509 subject or an OSCARS login
     * @param qualifier the domain of the Subject (an @ symbol will be appended by this method)
     */
    public void setSubject(String name, String qualifier) {
        SubjectBuilder subjBuilder = new SubjectBuilder();
        NameIDBuilder nameBuilder = new NameIDBuilder();
        SubjectConfirmationBuilder scBuilder = new SubjectConfirmationBuilder();
        
        //build subject
        Subject subject = subjBuilder.buildObject();
        
        //build NameID
        NameID nameId  = nameBuilder.buildObject();
        if(name.split(",").length > 1){
            nameId.setFormat(NameID.X509_SUBJECT);
        }else{
            nameId.setFormat(OSCARS_LOGIN_SUBJECT);
            nameId.setNameQualifier(qualifier);
        }
        
        nameId.setValue(name);
        subject.setNameID(nameId);
        
        //Always set SubjectConformation to SENDER-VOUCHES
        SubjectConfirmation subjectConfirmation = scBuilder.buildObject();
        subjectConfirmation.setMethod(SAMLSubject.CONF_SENDER_VOUCHES);
        List<SubjectConfirmation> subjConfirmList = subject.getSubjectConfirmations();
        subjConfirmList.add(subjectConfirmation);
        
        this.subject = subject;
    }
    
    /**
     * Sets a subject given a custom SAML subject
     * 
     * @param subject the subject to set
     */
    public void setSubject(Subject subject) {
        this.subject = subject;
    }
    /**
     * @return A SAML NameID containing a URL where the Subject can be verified.
     */
    public NameID getSubjectAuthentication() {
        return this.subjectAuthentication;
    }
    
    /**
     * Sets the SubjectAuthentication NameID given a URL as a string.
     * 
     * @param url the URL where the subject can be verified
     */
    public void setSubjectAuthentication(String url) {
        NameIDBuilder nameBuilder = new NameIDBuilder();

        NameID nameId  = nameBuilder.buildObject();
        nameId.setValue(url);
        this.subjectAuthentication = nameId;
    }
    
    /**
     * Sets the SubjectAuthentication field given a custom NameID
     * 
     * @param subjectAuthentication the SubjectAuthentication NameID to set.
     */
    public void setSubjectAuthentication(NameID subjectAuthentication) {
        this.subjectAuthentication = subjectAuthentication;
    }
    /**
     * @return SAML assertions about the subject
     */
    public List<Assertion> getAssertions() {
        return this.assertions;
    }
    /**
     * Sets SAML assertions about the subject
     * 
     * @param SAML assertions about the subject
     */
    public void setAssertions(List<Assertion> assertions) {
        this.assertions = assertions;
    }
    
    /**
     * Marshalls all the SAML types to W3C DOM Objects. Method
     * is automatically called in client classes such as Axis2PolicyClient.
     * 
     * @throws MarshallingException
     */
    public void marshall() throws MarshallingException{
        //Marshall subject to XML
        if(this.subject != null){
            SubjectMarshaller subjMarshaller = new SubjectMarshaller();
            this.subject.setDOM(subjMarshaller.marshall(this.subject));
        }
        
        //Marshall SubjectAuthentication
        if(this.subjectAuthentication != null){
            NameIDMarshaller nameIdMarshaller = new NameIDMarshaller();
            this.subjectAuthentication.setDOM(nameIdMarshaller.marshall(this.subjectAuthentication));
        }
        
        //Marshall Asserstions
        if(this.assertions != null && !this.assertions.isEmpty()){
            AssertionMarshaller assertionMarshaller = new AssertionMarshaller();
            for(Assertion assertion : this.assertions){
                assertionMarshaller.marshall(assertion);
            }
        }
    }
}
