package net.es.oscars.policy.rmi;

import java.rmi.RemoteException;
import java.util.List;
import java.util.Properties;

import net.es.oscars.PropHandler;
import net.es.oscars.policy.PolicyCore;
import net.es.oscars.policy.PolicyManager;
import net.es.oscars.policy.client.CheckPolicyRequest;
import net.es.oscars.policy.client.CheckPolicyResponse;
import net.es.oscars.policy.client.DOMRequester;
import net.es.oscars.policy.client.PolicyClientException;
import net.es.oscars.policy.client.Requester;
import net.es.oscars.policy.client.SAMLBootstrapper;
import net.es.oscars.policy.config.PolicyFileParser;
import net.es.oscars.rmi.BaseRmiServer;

import org.apache.log4j.Logger;
import org.hibernate.Session;
import org.opensaml.saml2.core.Assertion;
import org.opensaml.saml2.core.NameID;
import org.opensaml.saml2.core.Subject;
import org.opensaml.saml2.core.impl.AssertionUnmarshaller;
import org.opensaml.saml2.core.impl.NameIDUnmarshaller;
import org.opensaml.saml2.core.impl.SubjectUnmarshaller;
import org.opensaml.xml.io.UnmarshallingException;
import org.w3c.dom.Element;

/**
 * RMI server that accepts requests for the Policy Service core.
 * 
 * @author Andrew Lake (alake@internet2.edu)
 */
public class PolicyRmiServer extends BaseRmiServer implements PolicyRmiInterface{
    /** Static remote object so that GarbageCollector doesn't delete it */
    public static PolicyRmiServer staticObject;
    public PolicyManager policyMgr;
    
    public PolicyRmiServer(){
        this.log = Logger.getLogger(this.getClass());
        this.policyMgr = new PolicyManager();
    }
    
    /**
     * init
     *   By default initializes the RMI server registry to listen on port 1099 (the default)
     *   the RMI server to listen on a random port, and both to listen only on the loopback
     *   interface. These values can be overidden by oscars.properties.
     *   Setting the serverIpaddr to localhost will allow access from remote hosts and
     *   invalidate our security assumptions.
     *
     * @throws remoteException
     */
    public void init() throws RemoteException {
        this.log.debug("PolicyRmiServer.init().start");
        PolicyRmiServer.staticObject = this;
        
        PropHandler propHandler = new PropHandler("oscars.properties");
        Properties props = propHandler.getPropertyGroup("rmi.policy", true);
        this.setProperties(props);
        this.serviceName="Policy RMI Server";
        this.rmiServerName=PolicyRmiInterface.registeredServerName;
        try {
            SAMLBootstrapper.bootstrap();
        } catch (Exception e) {
            throw new RemoteException(e.getMessage());
        }
        super.init(staticObject);
    }

    /**
     * shutdown
     */
    public void shutdown() {
        super.shutdown(staticObject);
    }
    
    /**
     * Returns whether an action on a resource(s) is allowed by requester(s)
     * 
     * @param request the request on which to check policy
     * @return the result of the policy check
     * @throws RemoteException
     */
    public CheckPolicyResponse checkPolicy(CheckPolicyRequest<DOMRequester,Element[]> request) throws RemoteException{
        this.log.debug("checkPolicy.start");
        CheckPolicyRequest<Requester,Element[]> samlRequest = null;
        CheckPolicyResponse response = null;
        Session session = PolicyCore.getInstance().getPolicySession();
        session.beginTransaction();
        try {
            samlRequest = this.toSAMLRequest(request);
            response = this.policyMgr.checkPolicy(samlRequest);
            session.getTransaction().commit();
        } catch (Exception e) {
           session.getTransaction().rollback();
           e.printStackTrace();
           throw new RemoteException(e.toString());
        }
        
        this.log.debug("checkPolicy.end");
        return response;
    }
    
    /**
     * Converts an incoming checkPolicy request that uses W3C DOM types to SAML2
     * types. Requests come in using DOM types because they are Serializable and
     * SAML2 types are not. 
     * 
     * @param request the request containing DOM types
     * @return the converted request using SAML2 types
     * @throws UnmarshallingException
     * @throws PolicyClientException
     */
    private CheckPolicyRequest<Requester, Element[]> toSAMLRequest(
            CheckPolicyRequest<DOMRequester, Element[]> request) throws UnmarshallingException, PolicyClientException {
        this.log.debug("toSAMLRequest.start");
        CheckPolicyRequest<Requester, Element[]> samlRequest = new CheckPolicyRequest<Requester, Element[]>();
        SubjectUnmarshaller subjUnmarshaller = new SubjectUnmarshaller();
        NameIDUnmarshaller nameUnmarshaller = new NameIDUnmarshaller();
        AssertionUnmarshaller assertUnmarshaller = new AssertionUnmarshaller();
        for(DOMRequester domRequester : request.getRequesters()){
            Requester requester = new Requester();
            //Set Subject
            Subject subject = (Subject) subjUnmarshaller.unmarshall(domRequester.getSubject());
            requester.setSubject(subject);
            //Set Subject Authentication
            if(domRequester.getSubjectAuthentication() != null){
                NameID subjAuth = (NameID)nameUnmarshaller.unmarshall(domRequester.getSubjectAuthentication());
                requester.setSubjectAuthentication(subjAuth);
            }
            //Set Assertions
            List<Element> domAsserts = domRequester.getAssertions();
            if(domAsserts != null){
                for(Element domAssert : domAsserts){
                       Assertion assertion = (Assertion)assertUnmarshaller.unmarshall(domAssert);
                       requester.getAssertions().add(assertion);
                }
            }
            samlRequest.getRequesters().add(requester);
        }
        samlRequest.setAction(request.getAction());
        samlRequest.setResources(request.getResources());
        
        this.log.debug("toSAMLRequest.end");
        return samlRequest;
    }

    public void loadPolicy(String filename, Boolean verbose) throws RemoteException {
        try {
            PolicyFileParser fileParser = new PolicyFileParser(filename, verbose);
            fileParser.parse();
        } catch (Exception e) {
            this.log.error(e.getMessage());
            e.printStackTrace();
           throw new RemoteException(e.getMessage());
        }
        
    }
}
